@@ -1,4 +1,4 @@
-/*  $Id: pl-comp.c,v 1.40 1996/09/06 09:33:44 jan Exp $
+/*  $Id: pl-comp.c,v 1.43 1996/10/30 17:32:31 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -94,6 +94,8 @@ struct code_info codeTable[] = {
   CODE(C_SOFTIF,	"c_softif",	2, 0),
   CODE(C_SOFTCUT,	"c_softcut",	1, 0),
 #endif
+  CODE(I_EXITFACT,	"i_exitfact",	0, 0),
+  CODE(D_BREAK,		"d_break",	0, 0),
 /*List terminator */
   CODE(0,		NULL,		0, 0)
 };
@@ -158,10 +160,6 @@ BUGS:	Currently there are three  places were all the VM instructions
 	should  be merged.  For  now, be very carefull  if you add  or
 	delete a VM instruction.
 
-	Be carefull: pl-wam.o must be loaded in low addresses!
-		     make sure you have VM that lets you start the
-		     text addresses close to zero!
-
 NOTE:	If the assert() fails, look at pl-wam.c: VMI(C_NOT, ... for
 	more information.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
@@ -631,7 +629,6 @@ Allocate the clause and fill initialise the field we already know.
   clause = (Clause) allocHeap(sizeof(struct clause));
   clause->flags      = 0;
   clause->code_size  = 0;
-  clause->subclauses = 0;
   clause->procedure  = proc;
   clause->source_no  = clause->line_no = 0;
 
@@ -687,8 +684,11 @@ Now compile the body.
   if ( body && *body != ATOM_true )
   { Output_0(&ci, I_ENTER);
     compileBody(body, I_DEPART, &ci);
+    Output_0(&ci, I_EXIT);
+  } else
+  { set(clause, UNIT_CLAUSE);		/* fact (for decompiler) */
+    Output_0(&ci, I_EXITFACT);
   }
-  Output_0(&ci, I_EXIT);
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Reset all variables we initialised to the variable analysis  functor  to
@@ -793,7 +793,7 @@ compileBody(register Word body, code call, register compileInfo *ci)
 	tc_or = PC(ci);
 	TRY( compileBody(argTermP(*a0, 0), I_CALL, ci) );	
 	Output_1(ci, hard ? C_CUT : C_SOFTCUT, var);
-	TRY( compileBody(argTermP(*a0, 1), I_CALL, ci) );	
+	TRY( compileBody(argTermP(*a0, 1), call, ci) );	
 	balanceVars(valt1, valt2, ci);
 	Output_1(ci, C_JMP, (code)0);
 	tc_jmp = PC(ci);
@@ -852,7 +852,6 @@ compileBody(register Word body, code call, register compileInfo *ci)
   }
 
   TRY( compileSubClause(body, call, ci) );
-  ci->clause->subclauses++;
 
   succeed;
 }
@@ -1525,6 +1524,59 @@ pl_redefine_system_predicate(term_t pred)
 		*          DECOMPILER           *
 		*********************************/
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+decompileArg1()  is  a  simplified  version   of  decompileHead().   Its
+function is to extract the relevant   information  for (re)computing the
+index information for indexing on the   first argument (the 99.9% case).
+See reindexClause().
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+int
+arg1Key(Clause clause, word *key)
+{ Code PC = clause->codes;
+
+  for(;;)
+  { code c = decode(*PC++);
+
+  again:
+    switch(c)
+    { case H_FUNCTOR:
+      case H_RFUNCTOR:
+	*key = ((FunctorDef)*PC)->functor;
+        succeed;
+      case H_CONST:
+	*key = *PC;
+	succeed;
+      case H_NIL:
+	*key = ATOM_nil;
+        succeed;
+      case H_LIST:
+      case H_RLIST:
+	*key = FUNCTOR_dot2->functor;
+        succeed;
+      case H_INTEGER:
+      case H_FLOAT:			/* tbd */
+      case H_INDIRECT:
+      case H_FIRSTVAR:
+      case H_VAR:
+      case H_VOID:
+      case I_EXITFACT:
+      case I_EXIT:			/* fact */
+      case I_ENTER:			/* fix H_VOID, H_VOID, I_ENTER */
+	fail;
+      case I_NOP:
+	continue;
+      case D_BREAK:
+        c = decode(replacedBreak(PC-1));
+	goto again;
+      default:
+	assert(0);
+        fail;
+    }
+  }
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 The decompiler is rather straightforwards.  First it  will  construct  a
 term  with  variables  for  the  head  and an array of variables for all
@@ -1546,58 +1598,13 @@ typedef struct
   Word	 argp;				/* argument pointer */
   int	 nvars;				/* size of var block */
   term_t *variables;			/* variable table */
+  term_t bindings;			/* [Offset = Var, ...] */
 } decompileInfo;
 
 forwards bool	decompile_head(Clause, term_t, decompileInfo *);
 forwards bool	decompileBody(decompileInfo *, code, Code);
 forwards void	build_term(FunctorDef, decompileInfo *);
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-decompileArg1()  is  a  simplified  version   of  decompileHead().   Its
-function is to extract the relevant   information  for (re)computing the
-index information for indexing on the   first argument (the 99.9% case).
-See reindexClause().
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-int
-arg1Key(Clause clause, word *key)
-{ Code APC = clause->codes;
-
-  for(;;)
-  { switch(decode(*APC++))
-    { case H_FUNCTOR:
-      case H_RFUNCTOR:
-	*key = ((FunctorDef)XR(*APC))->functor;
-        succeed;
-      case H_CONST:
-	*key = XR(*APC);
-	succeed;
-      case H_NIL:
-	*key = ATOM_nil;
-        succeed;
-      case H_LIST:
-      case H_RLIST:
-	*key = FUNCTOR_dot2->functor;
-        succeed;
-      case H_INTEGER:
-      case H_FLOAT:			/* tbd */
-      case H_INDIRECT:
-      case H_FIRSTVAR:
-      case H_VAR:
-      case H_VOID:
-      case I_EXIT:			/* fact */
-      case I_ENTER:			/* fix H_VOID, H_VOID, I_ENTER */
-	fail;
-      case I_NOP:
-	continue;
-      default:
-	assert(0);
-        fail;
-    }
-  }
-}
-
-
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 decompileHead()  is  public  as  it  is   needed  to  update  the  index
 information for clauses if this changes   when  the predicate is already
@@ -1622,6 +1629,7 @@ decompileHead(Clause clause, term_t head)
 { decompileInfo di;
   di.nvars     = VAROFFSET(1) + clause->prolog_vars;
   di.variables = alloca(di.nvars * sizeof(term_t));
+  di.bindings  = 0;
 
   return decompile_head(clause, head, &di);
 }
@@ -1660,13 +1668,28 @@ decompile_head(Clause clause, term_t head, decompileInfo *di)
   int pushed = 0;
   Definition def = clause->procedure->definition;
 
-  { int m, l;
-    term_t *p;
+  if ( di->bindings )
+  { term_t *p = &di->variables[VAROFFSET(0)];
+    term_t tail = PL_copy_term_ref(di->bindings);
+    term_t head = PL_new_term_ref();
+    int n;
 
-    l = VAROFFSET(0);
-    m = VAROFFSET(clause->prolog_vars) - l; /* # prolog vars (+arguments) */
-    for(p = di->variables + l; m-- > 0;)
-      *p++ = PL_new_term_ref();
+    for(n=0; n<clause->prolog_vars; n++)
+    { p[n] = PL_new_term_ref();
+
+      if ( !PL_unify_list(tail, head, tail) ||
+	   !PL_unify_term(head, PL_FUNCTOR, FUNCTOR_equals2,
+			  	    PL_INTEGER, n,
+			            PL_TERM, p[n]) )
+	fail;
+    }
+    TRY(PL_unify_nil(tail));
+  } else
+  { term_t *p = &di->variables[VAROFFSET(0)];
+    int n;
+
+    for(n=0; n<clause->prolog_vars; n++)
+      p[n] = PL_new_term_ref();
   }
 
   argp  = PL_new_term_ref();
@@ -1681,13 +1704,19 @@ decompile_head(Clause clause, term_t head, decompileInfo *di)
 #define NEXTARG { next_arg_ref(argp); if ( !pushed ) argn++; }
 
   for(;;)
-  { switch(decode(*PC++))
+  { code c = decode(*PC++);
+
+  again:
+    switch(c)
     { case I_NOP:
-	  continue;
+	continue;
+      case D_BREAK:
+	c = decode(replacedBreak(PC-1));
+        goto again;
       case H_NIL:
-	  TRY(PL_unify_nil(argp));
-          NEXTARG;
-	  continue;
+	TRY(PL_unify_nil(argp));
+        NEXTARG;
+        continue;
       case H_INDIRECT:
         { word copy = globalIndirectFromCode(&PC);
 	  TRY(_PL_unify_atomic(argp, copy));
@@ -1762,6 +1791,7 @@ decompile_head(Clause clause, term_t head, decompileInfo *di)
 	  if ( !pushed )
 	    argn++;
 	  continue;
+      case I_EXITFACT:
       case I_EXIT:			/* fact */
       case I_ENTER:			/* fix H_VOID, H_VOID, I_ENTER */
 	{ assert(argn <= arity);
@@ -1786,21 +1816,22 @@ decompile_head(Clause clause, term_t head, decompileInfo *di)
 			  storage(w) == STG_INLINE) ? valInt(w) : -1)
 
 bool
-decompile(Clause clause, term_t term)
+decompile(Clause clause, term_t term, term_t bindings)
 { decompileInfo dinfo;
   decompileInfo *di = &dinfo;
   Word body;
 
   di->nvars     = VAROFFSET(1) + clause->prolog_vars;
   di->variables = alloca(di->nvars * sizeof(term_t));
+  di->bindings  = bindings;
 
 #ifdef O_RUNTIME
   if ( false(clause->procedure->definition, DYNAMIC) )
     fail;
 #endif
 
-  if ( clause->subclauses == 0 )	/* fact */
-  { return decompileHead(clause, term);
+  if ( true(clause, UNIT_CLAUSE) )	/* fact */
+  { return decompile_head(clause, term, di);
   } else
   { term_t a = PL_new_term_ref();
 
@@ -1870,11 +1901,22 @@ static bool
 decompileBody(register decompileInfo *di, code end, Code until)
 { int nested = 0;		/* nesting in FUNCTOR ... POP */
   int pushed = 0;		/* Subclauses pushed on the stack */
-  int op;
+  code op;
 
-  for(; decode(*PC) != end && PC != until; )
-  { switch((op=decode(*PC++)))
-    {   case A_ENTER:
+  while( PC != until )
+  { op = decode(*PC++);
+
+  again:
+    if ( op == end )
+    { PC--;
+      break;
+    }
+
+    switch( op )
+    {   case D_BREAK:	    op = decode(replacedBreak(PC-1));
+			    goto again;
+	  
+        case A_ENTER:
         case I_NOP:	    continue;
 	case B_CONST:
 			    *ARGP++ = XR(*PC++);
@@ -2182,7 +2224,7 @@ unify_definition(term_t head, Definition def, term_t thehead)
 
 
 word
-pl_clause(term_t p, term_t term, term_t ref, word h)
+pl_clause4(term_t p, term_t term, term_t ref, term_t bindings, word h)
 { Procedure proc;
   Definition def;
   Clause clause;
@@ -2202,7 +2244,7 @@ pl_clause(term_t p, term_t term, term_t ref, word h)
     if ( !inCore(clause) || !isClause(clause) )
       return warning("clause/3: Invalid reference");
 	
-    if ( !decompile(clause, term) )
+    if ( !decompile(clause, term, bindings) )
       fail;
 
     proc = clause->procedure;
@@ -2252,7 +2294,7 @@ pl_clause(term_t p, term_t term, term_t ref, word h)
       if ( !(cref = findClause(cref, argv, def, &det)) )
 	fail;
   
-      if ( !decompile(cref->clause, term) )
+      if ( !decompile(cref->clause, term, bindings) )
 	continue;
       if ( !PL_unify_pointer(ref, cref->clause) )
 	continue;
@@ -2268,6 +2310,12 @@ pl_clause(term_t p, term_t term, term_t ref, word h)
 }
 
 
+word
+pl_clause(term_t p, term_t term, term_t ref, word h)
+{ return pl_clause4(p, term, ref, 0, h);
+}
+
+
 typedef struct
 { ClauseRef clause;			/* pointer to the clause */
   int       index;			/* nth-1 index */
@@ -2436,6 +2484,9 @@ pl_xr_member(term_t ref, term_t term, word h)
     { bool rval = FALSE;
       code op = decode(*PC++);
       
+      if ( op == D_BREAK )
+	op = decode(replacedBreak(PC-1));
+
       switch(codeTable[op].argtype)
       { case CA1_PROC:
 	{ Procedure proc = (Procedure) *PC;
@@ -2535,11 +2586,21 @@ wamListClause(Clause clause)
   ep = bp + clause->code_size;
 
   while( bp < ep )
-  { code op = decode(*bp++);
-    CodeInfo ci = &codeTable[op];
+  { code op = decode(*bp);
+    CodeInfo ci;
     int n = 0;
+    int isbreak;
 
-    Putf("%4d %s", bp - 1 - clause->codes, ci->name);
+    if ( op == D_BREAK )
+    { op = decode(replacedBreak(bp));
+      isbreak = TRUE;
+    } else
+      isbreak = FALSE;
+
+    ci = &codeTable[op];
+
+    Putf("%4d %s", bp - clause->codes, ci->name);
+    bp++;
 
     switch(op)
     { case B_FIRSTVAR:
@@ -2627,6 +2688,9 @@ wamListClause(Clause clause)
 	  Putf("%s%d", n == 0 ? " " : ", ", *bp++);
     }
 
+    if ( isbreak )
+      Putf(" *break*");
+
     Putf("\n");
   }
 }
@@ -2645,3 +2709,510 @@ pl_wam_list(term_t ref)
   succeed;
 }
 
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+$fetch_vm(+Clause, +Offset, -NextOffset, -Instruction)
+	fetches the virtual machine instruction at the indicated position
+	and return NextOffset with the offset of the next instruction, or
+	[] if there is no next instruction.  Instruction is unified with
+	a descriptive term of the instruction, but for now only with the
+	name of the instruction.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+word
+pl_fetch_vm(term_t ref, term_t offset, term_t noffset, term_t instruction)
+{ Clause clause;
+  int pcoffset;
+  Code PC;
+  code op;
+  CodeInfo ci;
+
+  if ( !PL_get_pointer(ref, (void **)&clause) ||
+       !inCore(clause) || !isClause(clause) ||
+       !PL_get_integer(offset, &pcoffset) ||
+       pcoffset < 0 || pcoffset >= clause->code_size )
+    return warning("$fetch_vm/4: instantiation fault");
+
+  PC = clause->codes + pcoffset;
+  op = decode(*PC);
+  if ( op == D_BREAK )
+    op = decode(replacedBreak(PC));
+  ci = &codeTable[op];
+  
+  pcoffset = pcoffset + 1 + ci->arguments;
+
+  if ( PL_unify_integer(noffset, pcoffset) &&
+       PL_unify_atom_chars(instruction, ci->name) )
+    succeed;
+
+  fail;
+}
+
+
+
+		 /*******************************
+		 *     SOURCE LEVEL DEBUGGER	*
+		 *******************************/
+
+static Code
+find_code1(Code PC, code fop, code ctx)
+{ for(;;)
+  { code op = decode(*PC++);
+
+    if ( op == D_BREAK )
+      op = decode(replacedBreak(PC-1));
+
+    if ( fop == op && ctx == *PC )
+      return &PC[-1];
+    assert(op != I_EXIT);
+
+    PC += codeTable[op].arguments;
+  }
+}
+
+
+static Code
+find_code0(Code PC, code fop)
+{ for(;;)
+  { code op = decode(*PC++);
+
+    if ( op == D_BREAK )
+      op = decode(replacedBreak(PC-1));
+    if ( fop == op )
+      return &PC[-1];
+    assert(op != I_EXIT);
+
+    PC += codeTable[op].arguments;
+  }
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+$clause_term_position(+ClauseRef, +PCoffset, -TermPos)
+	Find the term-location of the call that ends in the given PC offset.
+	The term-position is a list of argument-numbers one has to use from
+	the clause-term to find the subterm that sets up the goal.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+/*
+#undef DEBUG
+#define DEBUG(l, g) g
+*/
+
+static int
+add_node(term_t tail, int n)
+{ term_t h = PL_new_term_ref();
+  int rval;
+
+  rval = PL_unify_list(tail, h, tail) && PL_unify_integer(h, n);
+  PL_reset_term_refs(h);
+
+  DEBUG(1, Sdprintf("Added %d\n", n));
+
+  return rval;
+}
+
+
+static void
+add_1_if_not_at_end(Code PC, Code end, term_t tail)
+{ while(PC < end && decode(*PC) == C_VAR )
+    PC += 2;
+
+  if ( PC != end )
+    add_node(tail, 1);
+}
+
+
+
+word
+pl_clause_term_position(term_t ref, term_t pc, term_t locterm)
+{ Clause clause;
+  int pcoffset;
+  Code PC, loc, end;
+  term_t tail = PL_copy_term_ref(locterm);
+
+  if ( !PL_get_pointer(ref, (void **)&clause) ||
+       !inCore(clause) || !isClause(clause) ||
+       !PL_get_integer(pc, &pcoffset) ||
+       pcoffset < 0 || pcoffset > clause->code_size )
+    return warning("$clause_location/3: invalid argument");
+
+  PC = clause->codes;
+  loc = &PC[pcoffset];
+  end = &PC[clause->code_size - 1];	/* forget the final I_EXIT */
+
+  while( PC < loc )
+  { code op = decode(*PC++);
+    CodeInfo ci;
+
+    if ( op == D_BREAK )
+      op = decode(replacedBreak(PC-1));
+    ci = &codeTable[op];
+
+    switch(op)
+    { case I_ENTER:
+	if ( loc == PC )
+	{ add_node(tail, 1);
+
+	  return PL_unify_nil(tail);
+	}
+	add_node(tail, 2);
+	continue;
+      case I_EXIT:
+      case I_EXITFACT:
+	if ( loc == PC )
+	{ return PL_unify_nil(tail);
+	}
+        continue;
+    { Code endloc;
+      case C_OR:			/* C_OR <jmp1> <A> C_JMP <jmp2> <B> */
+      { Code jmploc = PC + *PC++ + 1;
+
+	endloc = jmploc + jmploc[-1];
+
+	DEBUG(1, Sdprintf("jmp = %d, end = %d\n",
+			  jmploc - clause->codes, endloc - clause->codes));
+
+	if ( loc <= endloc )		/* loc is in the disjunction */
+	{ add_1_if_not_at_end(endloc, end, tail);
+
+	  if ( loc <= jmploc )		/* loc is in first branch */
+	  { add_node(tail, 1);
+	    end = jmploc-2;
+	    continue;
+	  }
+					/* loc is in second branch */
+	  add_node(tail, 2);
+	  PC = jmploc;
+	  end = endloc;
+	  continue;
+	}
+
+      after_construct:
+	add_node(tail, 2);		/* loc is after disjunction */
+	PC = endloc;
+	continue;
+      }
+      case C_NOT:		/* C_NOT <var> <jmp> <A> C_CUT <var>, C_FAIL */
+      { endloc = PC + PC[1] + 2;
+
+	DEBUG(1, Sdprintf("not: PC= %d, endloc = %d\n",
+			  PC - clause->codes, endloc - clause->codes));
+
+	if ( loc <= endloc )		/* in the \+ argument */
+	{ add_1_if_not_at_end(endloc, end, tail);
+
+	  add_node(tail, 1);
+	  PC += 2;
+	  end = endloc-3;		/* C_CUT <var>, C_FAIL */
+	  continue;
+	}
+
+	goto after_construct;
+      }
+      case C_SOFTIF:
+      case C_IFTHENELSE:	/* C_IFTHENELSE <var> <jmp1> */
+				/* <IF> C_CUT <THEN> C_JMP <jmp2> <ELSE> */
+      { Code elseloc = PC + PC[1] + 2;
+	code cut = (op == C_IFTHENELSE ? C_CUT : C_SOFTCUT);
+
+	endloc = elseloc + elseloc[-1];
+
+	DEBUG(1, Sdprintf("else = %d, end = %d\n",
+			  elseloc - clause->codes, endloc - clause->codes));
+
+	if ( loc <= endloc )
+	{ add_1_if_not_at_end(endloc, end, tail);
+
+	  if ( loc <= elseloc )		/* a->b */
+	  { Code cutloc = find_code1(&PC[2], cut, PC[0]);
+
+	    DEBUG(1, Sdprintf("cut at %d\n", cutloc - clause->codes));
+	    add_node(tail, 1);
+	    
+	    if ( loc <= cutloc )	/* a */
+	    { add_node(tail, 1);
+	      end = cutloc;
+	      PC = &PC[2];
+	    } else			/* b */
+	    { add_node(tail, 2);
+	      PC = cutloc + 2;
+	      end = elseloc-2;
+	    }    
+	    DEBUG(1, Sdprintf("end = %d\n", end - clause->codes));
+	    continue;
+	  }
+					/* c */
+	  add_node(tail, 2);
+	  PC = elseloc;
+	  end = endloc;
+	  continue;
+	}
+
+	goto after_construct;
+      }
+      case C_MARK:		/* A -> B */
+				/* C_MARK <var> <A> C_CUT <var> <B> C_END */
+      { Code cutloc = find_code1(&PC[1], C_CUT, PC[0]);
+	
+	endloc = find_code0(cutloc+2, C_END);
+
+	if ( loc <= endloc )
+	{ add_1_if_not_at_end(endloc, end, tail);
+
+	  if ( loc <= cutloc )		/* a */
+	  { add_node(tail, 1);
+
+	    PC += 1;
+	    end = cutloc;
+	  } else			/* b */
+	  { add_node(tail, 2);
+	    PC = cutloc+2;
+	    end = endloc;
+	  }
+
+	  continue;
+	}
+
+	goto after_construct;
+      }
+      }					/* closes the special constructs */
+      case I_CALL:
+      case I_DEPART:
+      case I_CUT:
+      case I_FAIL:
+      case I_TRUE:
+      case I_APPLY:
+      case I_USERCALL0:
+      case I_USERCALLN:
+      case I_CALL_FV0:
+      case I_CALL_FV1:
+      case I_CALL_FV2:
+	PC += ci->arguments;
+        if ( loc == PC )
+	{ add_1_if_not_at_end(PC, end, tail);
+
+	  return PL_unify_nil(tail);
+	}
+	add_node(tail, 2);
+	continue;
+      default:
+	PC += ci->arguments;
+    }
+  }
+
+  fail;					/* assert(0) */
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Generate (on backtracing), all  possible   break-points  of  the clause.
+Works in combination with pl_clause_term_position()   to  find the place
+for placing a break-point.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+word
+pl_break_pc(term_t ref, term_t pc, term_t nextpc, control_t h)
+{ Clause clause;
+  int offset;
+  Code PC, end;
+
+  switch( ForeignControl(h) )
+  { case FRG_CUTTED:
+      succeed;
+    case FRG_FIRST_CALL:
+      offset = 0;
+    case FRG_REDO:
+    default:
+      offset = ForeignContextInt(h);
+  }
+
+  
+  if ( !PL_get_pointer(ref, (void **)&clause) ||
+       !inCore(clause) || !isClause(clause) )
+    fail;
+  PC = clause->codes + offset;
+  end = clause->codes + clause->code_size;
+
+  while( PC < end )
+  { code op = decode(*PC);
+    Code next;
+
+    if ( op == D_BREAK )
+      op = decode(replacedBreak(PC));
+    next = PC + 1 + codeTable[op].arguments;
+
+    switch(op)
+    { case I_ENTER:
+      case I_EXIT:
+      case I_EXITFACT:
+      case I_CALL:
+      case I_DEPART:
+      case I_CUT:
+      case I_FAIL:
+      case I_TRUE:
+      case I_APPLY:
+      case I_USERCALL0:
+      case I_USERCALLN:
+      case I_CALL_FV0:
+      case I_CALL_FV1:
+      case I_CALL_FV2:
+	if ( PL_unify_integer(pc, PC-clause->codes) &&
+	     PL_unify_integer(nextpc, next-clause->codes) )
+	  ForeignRedoInt(next-clause->codes);
+    }
+
+    PC = next;
+  }
+
+  fail;
+}
+
+		 /*******************************
+		 *         BREAK-POINTS		*
+		 *******************************/
+
+static Table breakTable;
+
+typedef struct
+{ Clause	clause;			/* Associated clause */
+  int		offset;			/* Offset of the instruction */
+  code		saved_instruction;	/* The instruction saved */
+} break_point, *BreakPoint;
+
+
+static bool
+setBreak(Clause clause, int offset)
+{ Code PC = clause->codes + offset;
+
+  if ( !breakTable )
+    breakTable = newHTable(16);
+
+  if ( *PC != encode(D_BREAK) )
+  { BreakPoint bp = allocHeap(sizeof(break_point));
+
+    bp->clause = clause;
+    bp->offset = offset;
+    bp->saved_instruction = *PC;
+
+    addHTable(breakTable, PC, bp);
+    *PC = encode(D_BREAK);
+    set(clause, HAS_BREAKPOINTS);
+
+    callEventHook(PLEV_BREAK, clause, offset);
+    succeed;
+  }
+
+  fail;
+}
+
+
+static int
+clearBreak(Clause clause, int offset)
+{ Code PC = clause->codes + offset;
+  BreakPoint bp;
+  Symbol s;
+
+  if ( !breakTable || !(s=lookupHTable(breakTable, PC)) )
+    fail;
+
+  bp = (BreakPoint)s->value;
+  *PC = bp->saved_instruction;
+  freeHeap(bp, sizeof(*bp));
+  deleteSymbolHTable(breakTable, s);
+
+  callEventHook(PLEV_NOBREAK, clause, offset);
+  succeed;
+}
+
+
+void
+clearBreakPointsClause(Clause clause)
+{ if ( breakTable )
+  { Symbol s, n;
+
+    for( s = firstHTable(breakTable); s; s = n )
+    { BreakPoint bp = (BreakPoint)s->value;
+
+      n = nextHTable(breakTable, s);
+
+      if ( bp->clause == clause )
+	clearBreak(bp->clause, bp->offset);
+    }    
+  }
+
+  clear(clause, HAS_BREAKPOINTS);
+}
+
+
+code
+replacedBreak(Code PC)
+{ Symbol s;
+  BreakPoint bp;
+
+  if ( !breakTable || !(s=lookupHTable(breakTable, PC)) )
+    return (code) sysError("No saved instruction for break");
+  bp = (BreakPoint)s->value;
+
+  return bp->saved_instruction;
+}
+
+
+word
+pl_break_at(term_t ref, term_t pc, term_t set)
+{ Clause clause;
+  int offset;
+  atom_t a;
+
+  if ( !PL_get_pointer(ref, (void **)&clause) ||
+       !inCore(clause) || !isClause(clause) ||
+       !PL_get_atom(set, &a) ||
+       !PL_get_integer(pc, &offset) ||
+       offset < 0 || offset >= clause->code_size )
+    fail;
+
+  if ( a == ATOM_true )
+    return setBreak(clause, offset);
+  else
+    return clearBreak(clause, offset);
+}
+
+
+word
+pl_current_break(term_t ref, term_t pc, control_t h)
+{ Symbol symb;
+  
+  if ( !breakTable )
+    fail;
+
+  switch( ForeignControl(h) )
+  { case FRG_FIRST_CALL:
+      symb = firstHTable(breakTable);
+      break;
+    case FRG_REDO:
+      symb = ForeignContextPtr(h);
+      break;
+    case FRG_CUTTED:
+    default:
+      succeed;
+  }
+
+  for( ; symb; symb = nextHTable(moduleTable, symb) )
+  { BreakPoint bp = (BreakPoint) symb->value;
+
+    { fid_t cid = PL_open_foreign_frame();
+
+      if ( PL_unify_pointer(ref, bp->clause) &&
+	   PL_unify_integer(pc,  bp->offset) )
+      { if ( !(symb = nextHTable(moduleTable, symb)) )
+	  succeed;
+
+	ForeignRedoPtr(symb);
+      }
+
+      PL_discard_foreign_frame(cid);
+    }
+  }
+
+  fail;
+}
