@@ -1,4 +1,4 @@
-/*  $Id: pl-ext.c,v 1.52 1996/08/08 08:51:11 jan Exp $
+/*  $Id: pl-ext.c,v 1.55 1996/10/30 17:32:35 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -61,6 +61,7 @@ static struct foreign {
   FRG("skip",			1, pl_skip,			TRACE_ME),
   FRG("skip",			2, pl_skip2,			TRACE_ME),
   FRG("get_single_char",	1, pl_get_single_char,		TRACE_ME),
+  FRG("$rawtty",		1, pl_rawtty,			META),
   FRG("seeing",			1, pl_seeing,			TRACE_ME),
   FRG("telling",		1, pl_telling,			TRACE_ME),
   FRG("seen",			0, pl_seen,			TRACE_ME),
@@ -149,6 +150,7 @@ static struct foreign {
   FRG("number_chars",		2, pl_number_chars,		TRACE_ME),
   FRG("int_to_atom",		3, pl_int_to_atom,		TRACE_ME),
   FRG("$format_number",		3, pl_format_number,		TRACE_ME),
+  FRG("atom_prefix",		2, pl_atom_prefix,		TRACE_ME),
   FRG("concat",			3, pl_concat,			TRACE_ME),
   FRG("$concat_atom",		2, pl_concat_atom,		TRACE_ME),
   FRG("concat_atom",		3, pl_concat_atom3,		TRACE_ME),
@@ -176,11 +178,17 @@ static struct foreign {
 
   FRG("abolish",    		2, pl_abolish,		   META|TRACE_ME),
   FRG("$clause",    		3, pl_clause,	      NDET|META|TRACE_ME),
+  FRG("$clause",	        4, pl_clause4,	      NDET|META|TRACE_ME),
   FRG("nth_clause", 		3, pl_nth_clause,     NDET|META|TRACE_ME),
   FRG("retract",    		1, pl_retract,        NDET|META|TRACE_ME),
   FRG("retractall",		1, pl_retractall,	   META|TRACE_ME),
   FRG("$xr_member",		2, pl_xr_member,      NDET|META|TRACE_ME),
   FRG("$wam_list",		1, pl_wam_list,			TRACE_ME),
+  FRG("$fetch_vm",		4, pl_fetch_vm,			TRACE_ME),
+  FRG("$clause_term_position",	3, pl_clause_term_position,	TRACE_ME),
+  FRG("$break_pc",		3, pl_break_pc,		   NDET|TRACE_ME),
+  FRG("$break_at",		3, pl_break_at,			TRACE_ME),
+  FRG("$current_break",		2, pl_current_break,	   NDET|TRACE_ME),
   
   FRG("flag",			3, pl_flag,			TRACE_ME),
   FRG("recorda",		3, pl_recorda,			TRACE_ME),
@@ -202,11 +210,12 @@ static struct foreign {
   FRG("$get_predicate_attribute", 3, pl_get_predicate_attribute,META|TRACE_ME),
   FRG("$get_clause_attribute",  3, pl_get_clause_attribute,	TRACE_ME),
   FRG("$require",		1, pl_require,		   META|TRACE_ME),
-  FRG("$source_file",		2, pl_source_file,	   META|TRACE_ME),
+  FRG("source_file",		2, pl_source_file,    NDET|META|TRACE_ME),
   FRG("$time_source_file",	2, pl_time_source_file,	   NDET|TRACE_ME),
   FRG("$start_consult",		1, pl_start_consult,		TRACE_ME),
   FRG("$make_system_source_files",0,pl_make_system_source_files,TRACE_ME),
   FRG("$default_predicate",	2, pl_default_predicate,   META|TRACE_ME),
+  FRG("$clause_from_source",	3, pl_clause_from_source,	TRACE_ME),
 
   FRG("var",			1, pl_var,			TRACE_ME),
   FRG("nonvar",			1, pl_nonvar,			TRACE_ME),
