@@ -1,4 +1,4 @@
-/*  $Id: pl-funcs.h,v 1.68 1996/08/08 08:51:12 jan Exp $
+/*  $Id: pl-funcs.h,v 1.71 1996/10/30 17:32:37 jan Exp $
 
     Part of SWI-Prolog
     Designed and implemented by Jan Wielemaker
@@ -107,14 +107,23 @@ word		pl_record_clause(term_t term, term_t file, term_t ref);
 word		pl_redefine_system_predicate(term_t term);
 bool		decompileHead(Clause clause, term_t head);
 int		arg1Key(Clause clause, word *key);
-bool		decompile(Clause clause, term_t term);
+bool		decompile(Clause clause, term_t term, term_t bindings);
+word		pl_clause4(term_t p, term_t t, term_t ref, term_t b, word h);
 word		pl_clause(term_t p, term_t term, term_t ref, word h);
 word		pl_nth_clause(term_t p, term_t n, term_t ref, word h);
 word		pl_xr_member(term_t ref, term_t term, word h);
 void		wamListClause(Clause clause);
 word		pl_wam_list(term_t ref);
+word		pl_fetch_vm(term_t ref, term_t offset, term_t noffset,
+			    term_t instruction);
 int		unify_definition(term_t head, Definition def,
 				 term_t thehead);
+word		pl_clause_term_position(term_t ref, term_t pc, term_t locterm);
+word		pl_break_pc(term_t ref, term_t pc, term_t nextpc, control_t h);
+word		pl_break_at(term_t ref, term_t pc, term_t set);
+code		replacedBreak(Code PC);
+void		clearBreakPointsClause(Clause clause);
+word		pl_current_break(term_t ref, term_t pc, control_t h);
 
 /* pl-dump.c */
 word		saveProgram(term_t new);
@@ -157,6 +166,7 @@ void		protocol(char *s, int n);
 int		currentLinePosition(void);
 void		setCurrentSourceLocation(void);
 int		getSingleChar(void);
+word		pl_rawtty(term_t goal);
 bool		readLine(char *buffer);
 bool		LockStream(void);
 bool		UnlockStream(void);
@@ -276,6 +286,7 @@ word		pl_garbage_collect(term_t d);
 void		resetGC(void);
 Word		findGRef(int n);
 int		growStacks(LocalFrame fr, Code PC, int l, int g, int t);
+void		clearUninitialisedVarsFrame(LocalFrame, Code);
 
 /* pl-glob.c */
 word		pl_wildcard_match(term_t pattern, term_t string);
@@ -429,6 +440,7 @@ word		pl_name(term_t atom, term_t string);
 word		pl_atom_chars(term_t atom, term_t string);
 word		pl_number_chars(term_t number, term_t string);
 word		pl_atom_char(term_t atom, term_t chr);
+word		pl_atom_prefix(term_t atom, term_t prefix);
 word		pl_concat(term_t a1, term_t a2, term_t a3);
 word		pl_concat_atom(term_t list, term_t atom);
 word		pl_concat_atom3(term_t list, term_t sep, term_t atom);
@@ -499,13 +511,14 @@ word		pl_index(term_t pred);
 SourceFile	lookupSourceFile(atom_t name);
 void		addProcedureSourceFile(SourceFile sf, Procedure proc);
 word		pl_make_system_source_files(void);
-word		pl_source_file(term_t descr, term_t file);
-word		pl_time_source_file(term_t file, term_t t, word h);
+word		pl_source_file(term_t descr, term_t file, control_t h);
+word		pl_time_source_file(term_t file, term_t t, control_t h);
 word		pl_start_consult(term_t file);
 word		pl_default_predicate(term_t d1, term_t d2);
 Definition	autoImport(FunctorDef f, Module m);
 word		pl_require(term_t pred);
 word		pl_check_definition(term_t spec);
+word		pl_clause_from_source(term_t file, term_t line, term_t clause);
 
 /* pl-prof.c */
 void		stopItimer(void);
@@ -582,17 +595,20 @@ Table		newHTable(int size);
 void		destroyHTable(Table ht);
 Symbol		lookupHTable(Table ht, Void name);
 bool		addHTable(Table ht, Void name, Void value);
+void		deleteSymbolHTable(Table ht, Symbol s);
 Symbol		nextHTable(Table ht, Symbol s);
 Symbol		firstHTable(Table ht);
 void		clearHTable(Table ht);
 void		initTables();
 
 /* pl-trace.c */
-int		tracePort(LocalFrame frame, int port);
+int		tracePort(LocalFrame frame, LocalFrame bfr, int port, Code PC);
 void		backTrace(LocalFrame frame, int depth);
 word		pl_trace_continuation(term_t what);
 void		initTracer(void);
 void		resetTracer(void);
+int		tracemode(int new, int *old);
+int		debugmode(int new, int *old);
 word		pl_trace(void);
 word		pl_notrace(void);
 word		pl_tracing(void);
@@ -608,6 +624,7 @@ word		pl_debuglevel(term_t old, term_t new);
 word		pl_unknown(term_t old, term_t new);
 word		pl_prolog_current_frame(term_t fr);
 word		pl_prolog_frame_attribute(term_t fr, term_t key, term_t val);
+void		callEventHook(int ev, ...);
 
 /* pl-util.c */
 char		digitName(int n, bool small);
