@@ -1,4 +1,4 @@
-/*  $Id: pl-gc.c,v 1.33 1996/04/18 12:38:06 jan Exp $
+/*  $Id: pl-gc.c,v 1.36 1996/10/30 17:32:41 jan Exp $
 
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
@@ -143,7 +143,6 @@ Marking, testing marks and extracting values from GC masked words.
 forwards void		mark_variable(Word);
 forwards void		mark_foreign(void);
 forwards void		sweep_foreign(void);
-forwards void		clear_uninitialised(LocalFrame, Code);
 forwards QueryFrame	mark_environments(LocalFrame);
 forwards TrailEntry	mark_choicepoints(LocalFrame, TrailEntry);
 forwards void		mark_stacks(LocalFrame);
@@ -180,7 +179,7 @@ static long relocation_cells;	/* # relocation cells */
 static long relocated_cells;	/* # relocated cells */
 static long needs_relocation;	/* # cells that need relocation */
 static long local_marked;	/* # cells marked local -> global ptrs */
-#if O_SHIFT_STACKS || O_SECURE
+#if O_SHIFT_STACKS || O_SECURE || defined(O_MAINTENANCE)
 static long local_frames;	/* frame count for debugging */
 #endif
 #if O_SECURE
@@ -461,33 +460,34 @@ mark_foreign()
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-clear_uninitialised(LocalFrame fr, Code PC);
+clearUninitialisedVarsFrame(LocalFrame fr, Code PC);
 
 Assuming the clause associated will resume   execution  at PC, determine
 the variables that are not yet initialised and set them to be variables.
 This  avoids  the  garbage  collector    considering  the  uninitialised
 variables.
 
-There are 5 instructions  that  assume   their  argument  variable to be
-uninstantiated.  The instruction C_JMP is used   by if-then-else as well
-as \+/1 (converted to (cond -> fail  ;   true))  to jump over the `else'
-branch.  If we hit a C_JMP, we   will  have initialised all variables in
-the `cond -> if' part and  the variable balancing instructions guarantee
-the same variables will be initialised in the `else' brance.
-
 [Q] wouldn't it be better to track  the variables that *are* initialised
 and consider the others to be not?  Might   take more time, but might be
 more reliable and simpler.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static void
-clear_uninitialised(LocalFrame fr, Code PC)
+void
+clearUninitialisedVarsFrame(LocalFrame fr, Code PC)
 { if ( PC != NULL )
   { Code branch_end = NULL;
+    code c;
 
-    for( ; ; PC += (codeTable[decode(*PC)].arguments + 1))
-    { switch(decode(*PC))
-      { case I_EXIT:
+    for( ; ; PC += (codeTable[c].arguments + 1))
+    { c = decode(*PC);
+
+    again:
+      switch( c )
+      { case D_BREAK:
+	  c = decode(replacedBreak(PC));
+	  goto again;
+	case I_EXIT:
+	case I_EXITFACT:
 	  return;
 	case C_JMP:
 	  if ( PC >= branch_end )
@@ -510,7 +510,7 @@ clear_uninitialised(LocalFrame fr, Code PC)
 	  }
 	  break;
       }
-      if ( decode(*PC) > I_HIGHEST )
+      if ( c > I_HIGHEST )
 	sysError("GC: Illegal WAM instruction: %d", decode(*PC));
     }
   }
@@ -546,7 +546,7 @@ mark_environments(LocalFrame fr)
     DEBUG(3, Sdprintf("Marking [%ld] %s\n",
 		      levelFrame(fr), predicateName(fr->predicate)));
 
-    clear_uninitialised(fr, PC);
+    clearUninitialisedVarsFrame(fr, PC);
 
     slots   = (PC == NULL ? fr->predicate->functor->arity : slotsFrame(fr));
 #if O_SECURE
@@ -1168,7 +1168,7 @@ considerGarbageCollect(Stack s)
 #endif /* O_DYNAMIC_STACKS */
 
 
-#if O_SECURE
+#if O_SECURE || O_DEBUG || defined(O_MAINTENANCE)
 static bool
 scan_global(int marked)
 { Word current;
@@ -1246,9 +1246,10 @@ LocalFrame fr;
       return NULL;			/* from choicepoints only */
     set(fr, FR_MARKED);
     local_frames++;
-    clear_uninitialised(fr, PC);
+    clearUninitialisedVarsFrame(fr, PC);
 
     check_mark(&fr->mark);
+    assert(onStack(local, fr));
 
     DEBUG(3, Sdprintf("Check [%ld] %s:",
 		      levelFrame(fr),
@@ -1536,7 +1537,7 @@ update_environments(LocalFrame fr, Code PC, long ls, long gs, long ts)
     }
 
     if ( ls )				/* update variables */
-    { clear_uninitialised(fr, PC);
+    { clearUninitialisedVarsFrame(fr, PC);
 
       slots = (PC == NULL ? fr->predicate->functor->arity : slotsFrame(fr));
       sp = argFrameP(fr, slots);
