@@ -1,4 +1,4 @@
-/*  $Id: pl-incl.h,v 1.85 1996/09/06 09:33:51 jan Exp $
+/*  $Id: pl-incl.h,v 1.88 1996/10/30 17:32:44 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -522,7 +522,10 @@ codes.
 #define C_SOFTCUT	((code)74)		/* `Cut' of A *-> B ; C */
 #endif /*O_SOFTCUT*/
 
-#define I_HIGHEST	((code)74)		/* largest WAM code !!! */
+#define I_EXITFACT	((code)75)		/* exit from a fact */
+#define D_BREAK		((code)76)		/* Debugger break-point */
+
+#define I_HIGHEST	((code)76)		/* largest WAM code !!! */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Arithmetic comparison
@@ -571,7 +574,9 @@ Files and streams.  Don't change the numbers, or change FOPENMODE below.
 	  word SOrval; \
 	  if ( SOn < 0 ) fail; \
 	  Output = SOn; \
+	  LockStream(); \
 	  SOrval = goal; \
+	  UnlockStream(); \
 	  Output = SOout; \
 	  return SOrval; \
 	}
@@ -793,6 +798,9 @@ with one operation, it turns out to be faster as well.
 #define ISCASE			(0x00200000L) /* predicate */
 
 #define ERASED			(0x0001) /* clause */
+#define UNIT_CLAUSE		(0x0002) /* clause */
+#define HAS_BREAKPOINTS		(0x0004) /* clause */
+
 #define UNKNOWN			(0x0002) /* module */
 
 #define INLINE_F		(0x0001) /* functor */
@@ -919,7 +927,6 @@ struct clause
   Code		codes;		/* byte codes of clause */
   struct index	index;		/* index key of clause */
   short		code_size;	/* size of byte code array */
-  short		subclauses;	/* number of subclauses in body (decompiler) */
   short		variables;		/* # of variables for frame */
   short		prolog_vars;		/* # real Prolog variables */
   unsigned short	line_no;	/* Source line-number */
@@ -1207,6 +1214,18 @@ typedef struct
   } value;
 } opt_spec, *OptSpec;
 
+		 /*******************************
+		 *	      EVENTS		*
+		 *******************************/
+
+#define PLEV_ERASED	0		/* clause or record was erased */
+#define PLEV_DEBUGGING	1		/* changed debugging mode */
+#define PLEV_TRACING	2		/* changed tracing mode */
+#define PLEV_SPY	3		/* changed spypoint */
+#define PLEV_BREAK	4		/* a break-point was set */
+#define PLEV_NOBREAK	5		/* a break-point was cleared */
+
+
 		/********************************
 		*             STACKS            *
 		*********************************/
@@ -1449,6 +1468,7 @@ GLOBAL Procedure	PROCEDURE_garbage_collect0;
 GLOBAL Procedure	PROCEDURE_block3;
 GLOBAL Procedure	PROCEDURE_true0;
 GLOBAL Procedure	PROCEDURE_fail0;
+GLOBAL Procedure	PROCEDURE_event_hook1;
 
 extern struct code_info	codeTable[];
 
@@ -1466,11 +1486,15 @@ Tracer communication declarations.
 #define ACTION_IGNORE	3
 #define ACTION_AGAIN	4
 
-#define CALL_PORT	0x1		/* port masks */
-#define EXIT_PORT	0x2
-#define FAIL_PORT	0x4
-#define REDO_PORT	0x8
+#define CALL_PORT	0x01		/* port masks */
+#define EXIT_PORT	0x02
+#define FAIL_PORT	0x04
+#define REDO_PORT	0x08
 #define UNIFY_PORT	0x10
+#define BREAK_PORT	0x20
+#define CUT_CALL_PORT   0x40
+#define CUT_EXIT_PORT   0x80
+#define CUT_PORT	(CUT_CALL_PORT|CUT_EXIT_PORT)
 #define VERY_DEEP	10000000L	/* deep skiplevel */
 
 #define LONGATOM_CHECK	    0x1		/* read/1: error on long atoms */
@@ -1494,6 +1518,7 @@ GLOBAL struct debuginfo
   bool		showContext;		/* tracer shows context module */
   int		styleCheck;		/* source style checking */
   int		suspendTrace;		/* tracing is suspended now */
+  LocalFrame	retryFrame;		/* Frame to retry */
 } debugstatus;
 
 #define CHARESCAPE_FEATURE	0x01	/* handle \ in atoms */
