@@ -1,4 +1,4 @@
-/*  $Id: pl-main.c,v 1.48 1996/08/02 07:33:54 jan Exp $
+/*  $Id: pl-main.c,v 1.52 1996/11/11 16:27:26 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -103,8 +103,7 @@ findHome(char *symbols, char *def)
   }
 #endif
 
-  fatalError("Can't find Prolog Home Directory");
-  return (char *) NULL;
+  return NULL;
 }
 
 /*
@@ -142,10 +141,12 @@ proposeStartupFile(char *symbols)
     return store_string(state);
   }
 
-  Ssprintf(state, "%s/startup/startup.%s",
-	  systemDefaults.home, systemDefaults.arch);
-
-  return store_string(AbsoluteFile(state));
+  if ( systemDefaults.home )
+  { Ssprintf(state, "%s/startup/startup.%s",
+	     systemDefaults.home, systemDefaults.arch);
+    return store_string(AbsoluteFile(state));
+  } else
+    return store_string("pl.qlf");
 }
 
 
@@ -158,47 +159,65 @@ findState(char *symbols)
   if ( AccessFile(full, ACCESS_READ) )
     return full;
 
-  Ssprintf(state, "%s/startup/startup.%s",
-	  systemDefaults.home, systemDefaults.arch);
-  full = AbsoluteFile(state);
-  if ( AccessFile(full, ACCESS_READ) )
-    return store_string(full);
+  if ( systemDefaults.home )
+  { Ssprintf(state, "%s/startup/startup.%s",
+	     systemDefaults.home, systemDefaults.arch);
+    full = AbsoluteFile(state);
+    if ( AccessFile(full, ACCESS_READ) )
+      return store_string(full);
 
-  Ssprintf(state, "%s/startup/startup", systemDefaults.home);
-  full = AbsoluteFile(state);
-  if ( AccessFile(full, ACCESS_READ) )
-    return store_string(full);
+    Ssprintf(state, "%s/startup/startup", systemDefaults.home);
+    full = AbsoluteFile(state);
+    if ( AccessFile(full, ACCESS_READ) )
+      return store_string(full);
+  }
 
   return NULL;
 }
 
 
+#ifndef O_RUNTIME
 static void
 warnNoFile(char *file)
 { AccessFile(file, ACCESS_READ);	/* just to set errno */
 
   Sfprintf(Serror, "    no `%s': %s\n", file, OsError());
 }
-
+#endif
 
 static void
 warnNoState()
-{ char state[MAXPATHLEN];
+{
+#ifdef O_RUNTIME
+  Sfprintf(Serror, "[FATAL ERROR: Runtime system: can not find a state to run\n");
+  Sfprintf(Serror, "\tUsage: %s -x state\n", mainArgv[0]);
+  Sfprintf(Serror, "\t\twhere <state> is created using qsave_program/[1,2]\n");
+  Sfprintf(Serror, "\t\tin the development system]\n");
+#else
+  char state[MAXPATHLEN];
   char *full;
 
   Sfprintf(Serror, "[FATAL ERROR: Failed to find startup file\n");
-  warnNoFile(proposeStartupFile(NULL));
-  Ssprintf(state, "%s/startup/startup.%s",
-	  systemDefaults.home, systemDefaults.arch);
-  full = AbsoluteFile(state);
-  warnNoFile(full);
-  Ssprintf(state, "%s/startup/startup", systemDefaults.home);
-  full = AbsoluteFile(state);
-  warnNoFile(full);
+  full = proposeStartupFile(NULL);
+  if ( full )
+    warnNoFile(full);
+  if ( systemDefaults.home )
+  { Ssprintf(state, "%s/startup/startup.%s",
+	     systemDefaults.home, systemDefaults.arch);
+    full = AbsoluteFile(state);
+    warnNoFile(full);
+
+    Ssprintf(state, "%s/startup/startup", systemDefaults.home);
+    full = AbsoluteFile(state);
+    warnNoFile(full);
+  } else
+    Sfprintf(Serror, "    No home directory!\n");
+
   Sfprintf(Serror,
-	  "\nUse\n\t`%s -o startup-file -b boot/init.pl -c boot/load.pl'\n",
+	  "\nUse\n\t`%s -O -o startup-file -b boot/init.pl'\n",
 	  mainArgv[0]);
   Sfprintf(Serror, "\nto create one]\n");
+#endif
 
   Halt(1);
 }
@@ -270,10 +289,11 @@ startProlog(int argc, char **argv, char **env)
     systemDefaults.home	       = findHome(symbols,
 					  store_string(PrologPath(PLHOME,plp)));
 #ifdef O_XOS
-  { char buf[MAXPATHLEN];
-    _xos_limited_os_filename(systemDefaults.home, buf);
-    systemDefaults.home = store_string(buf);
-  }
+    if ( systemDefaults.home )
+    { char buf[MAXPATHLEN];
+      _xos_limited_os_filename(systemDefaults.home, buf);
+      systemDefaults.home = store_string(buf);
+    }
 #endif
 
     systemDefaults.startup     = store_string(PrologPath(DEFSTARTUP, plp));
@@ -551,9 +571,14 @@ runtime_vars()
 	  "PLARCH=\"%s\";\n"
 	  "PLLIBS=\"%s\";\n"
 	  "PLLDFLAGS=\"%s\";\n"
-	  "PLVERSION=\"%d\";\n",
+	  "PLVERSION=\"%d\";\n"
+#if defined(HAVE_DLOPEN) || defined(HAVE_SHL_LOAD)
+	  "PLSHARED=\"yes\";\n",
+#else
+	  "PLSHARED=\"no\";\n",
+#endif
 	  C_CC,
-	  systemDefaults.home,
+	  systemDefaults.home ? systemDefaults.home : "<no home>",
 	  ARCH,
 	  C_LIBS,
 	  C_LDFLAGS,
