@@ -1,4 +1,4 @@
-/*  $Id: pl-prims.c,v 1.47 1996/09/06 09:34:04 jan Exp $
+/*  $Id: pl-prims.c,v 1.49 1996/10/30 17:32:56 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -1448,13 +1448,28 @@ pl_atom_char(term_t atom, term_t chr)
 
 
 static bool
-isPrefix(register char *s, register char *q)
+isPrefix(char *s, char *q)		/* s is prefix of q */
 { while(*s && *s == *q)
     s++, q++;
 
   return *s == EOS;
 }
 
+
+word
+pl_atom_prefix(term_t atom, term_t prefix)
+{ char *a, *p;
+
+  PL_get_chars(atom,   &a, CVT_ATOMIC|BUF_RING);
+  PL_get_chars(prefix, &p, CVT_ATOMIC|BUF_RING);
+
+  if ( a && p )
+    return isPrefix(p, a) ? TRUE : FALSE;
+
+  return warning("atom_prefix/2: instantiation fault");
+}
+
+
 static word
 concat(term_t a1, term_t a2, term_t a3, int (*out)(term_t, const char *))
 { char *s1 = NULL, *s2 = NULL, *s3 = NULL;
@@ -1467,7 +1482,8 @@ concat(term_t a1, term_t a2, term_t a3, int (*out)(term_t, const char *))
 
   if (s1 && s2)
   { l1 = strlen(s1);
-    tmp = alloca(l1 + strlen(s2));
+    l2 = strlen(s2);
+    tmp = alloca(l1 + l2 + 1);
     strcpy(tmp, s1);
     strcpy(tmp+l1, s2);
     return (*out)(a3, tmp);
