@@ -1,4 +1,4 @@
-/*  $Id: pl-proc.c,v 1.43 1996/07/30 15:17:12 jan Exp $
+/*  $Id: pl-proc.c,v 1.47 1996/11/11 13:35:41 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -89,7 +89,7 @@ isCurrentProcedure(FunctorDef f, Module m)
 }
 
 bool
-isDefinedProcedure(register Procedure proc)
+isDefinedProcedure(Procedure proc)
 { if ( /* true(proc->definition, FOREIGN) || not needed; union */
        proc->definition->definition.clauses ||
        true(proc->definition, DYNAMIC) )
@@ -430,7 +430,11 @@ retractClauseProcedure(Procedure proc, Clause clause)
 	    def->lastClause = prev;
 	}
 
-	freeClauseRef(c);
+
+  	freeClauseRef(c);
+	if ( PROCEDURE_event_hook1 &&
+	     def != PROCEDURE_event_hook1->definition )
+	  callEventHook(PLEV_ERASED, clause);
 	freeClause(clause);
 	def->number_of_clauses--;
 
@@ -448,7 +452,10 @@ retractClauseProcedure(Procedure proc, Clause clause)
 
 void
 freeClause(Clause c)
-{ statistics.codes -= c->code_size;
+{ if ( true(c, HAS_BREAKPOINTS) )
+    clearBreakPointsClause(c);
+
+  statistics.codes -= c->code_size;
   freeHeap(c->codes, sizeof(code) * c->code_size);
   freeHeap(c, sizeof(struct clause));
 }
@@ -492,6 +499,8 @@ gcClausesDefinition(Definition def)
       }
 
       DEBUG(0, removed++);
+      if ( PROCEDURE_event_hook1 && def != PROCEDURE_event_hook1->definition )
+	callEventHook(PLEV_ERASED, c->clause);
       freeClause(c->clause);
       freeClauseRef(c);
     } else
@@ -815,7 +824,7 @@ pl_retract(term_t term, word h)
 
       { fid_t cid = PL_open_foreign_frame();
 
-	if ( decompile(cref->clause, cl) )
+	if ( decompile(cref->clause, cl, 0) )
 	{ retractClauseProcedure(proc, cref->clause);
 	  PL_close_foreign_frame(cid);	/* necessary? */
 	  if ( det == TRUE )
@@ -859,8 +868,8 @@ pl_retractall(term_t head)
   { bool det;
     Word argv;
 
-    if ( PL_is_compound(head) )
-    { argv = valTermRef(head);
+    if ( PL_is_compound(thehead) )
+    { argv = valTermRef(thehead);
       deRef(argv);
       argv = argTermP(*argv, 0);
     } else
@@ -1000,6 +1009,8 @@ pl_get_predicate_attribute(term_t pred,
   { return PL_unify_integer(value, def->hash_info?def->hash_info->buckets:0);
   } else if ( key == ATOM_references )
   { return PL_unify_integer(value, def->references);
+  } else if ( key == ATOM_number_of_clauses )
+  { return PL_unify_integer(value, def->number_of_clauses);
   } else if ( (att = attribute_mask(key)) )
   { return PL_unify_integer(value, (def->flags & att) ? 1 : 0);
   } else
@@ -1091,7 +1102,7 @@ reindexDefinition(Definition def)
       def->indexPattern = 0x0;
     } else
     { def->indexPattern = 0x1;
-      if ( (cannotindex * 8 < canindex) && (canindex > 5) )
+      if ( canindex > 5 && cannotindex <= 2 )
 	do_hash = canindex / 2;
     }
   }
@@ -1176,6 +1187,10 @@ pl_get_clause_attribute(term_t ref, term_t att, term_t value)
     
     if ( sf )
       return PL_unify_atom(value, sf->name);
+  } else if ( a == ATOM_fact )
+  { return PL_unify_atom(value,
+			 true(clause, UNIT_CLAUSE) ? ATOM_true
+			 			   : ATOM_false);
   }
 
   fail;
@@ -1265,24 +1280,70 @@ pl_make_system_source_files(void)
   succeed;
 }
 
+
 word
-pl_source_file(term_t descr, term_t file)
+pl_source_file(term_t descr, term_t file, control_t h)
 { Procedure proc;
   ClauseRef cref;
   SourceFile sf;
+  atom_t name;
+  ListCell cell;
+  
 
-  if ( !get_procedure(descr, &proc, 0, GP_FIND) ||
-       !proc->definition ||
-       true(proc->definition, FOREIGN) ||
-       !(cref = proc->definition->definition.clauses) ||
-       !(sf = indexToSourceFile(cref->clause->source_no)) )
+  if ( ForeignControl(h) == FRG_FIRST_CALL &&
+       !PL_is_variable(descr) )
+  { if ( !get_procedure(descr, &proc, 0, GP_FIND) ||
+	 !proc->definition ||
+	 true(proc->definition, FOREIGN) ||
+	 !(cref = proc->definition->definition.clauses) ||
+	 !(sf = indexToSourceFile(cref->clause->source_no)) )
+      fail;
+
+    return PL_unify_atom(file, sf->name);
+  }
+
+  if ( ForeignControl(h) == FRG_CUTTED )
+    succeed;
+
+  if ( !PL_get_atom(file, &name) ||
+       !(sf = lookupSourceFile(name)) )
     fail;
 
-  return PL_unify_atom(file, sf->name);
+  switch( ForeignControl(h) )
+  { case FRG_FIRST_CALL:
+      cell = sf->procedures;
+      break;
+    case FRG_REDO:
+      cell = ForeignContextPtr(h);
+      break;
+    default:
+      cell = NULL;
+      assert(0);
+  }
+  
+  for( ; cell; cell = cell->next )
+  { Procedure proc = cell->value;
+    Definition def = proc->definition;
+    fid_t cid = PL_open_foreign_frame();
+
+    if ( unify_definition(descr, def, 0) )
+    { PL_close_foreign_frame(cid);
+
+      if ( cell->next )
+	ForeignRedoPtr(cell->next);
+      else
+	succeed;
+    }
+
+    PL_discard_foreign_frame(cid);
+  }
+
+  fail;
 }
 
+
 word
-pl_time_source_file(term_t file, term_t time, word h)
+pl_time_source_file(term_t file, term_t time, control_t h)
 { SourceFile fr;
 
   switch( ForeignControl(h) )
@@ -1346,5 +1407,49 @@ pl_start_consult(term_t file)
     succeed;
   }
 
+  fail;
+}
+
+		 /*******************************
+		 *       DEBUGGER SUPPORT	*
+		 *******************************/
+
+word
+pl_clause_from_source(term_t file, term_t line, term_t clause)
+{ atom_t name;
+  SourceFile f;
+  int ln;
+  ListCell cell;
+  Clause c = NULL;
+
+  if ( !PL_get_atom(file, &name) ||
+       !(f = lookupSourceFile(name)) ||
+       !PL_get_integer(line, &ln) )
+    return warning("clause_from_source/3: instantiation fault");
+  
+
+  for(cell = f->procedures; cell; cell = cell->next)
+  { Procedure proc = cell->value;
+    Definition def = proc->definition;
+
+    if ( def && false(def, FOREIGN) )
+    { ClauseRef cref = def->definition.clauses;
+
+      for( ; cref; cref = cref->next )
+      { Clause cl = cref->clause;
+
+	if ( cl->source_no == f->index )
+	{ if ( ln >= cl->line_no )
+	  { if ( !c || c->line_no < cl->line_no )
+	      c = cl;
+	  }
+	}
+      }
+    }
+  }
+
+  if ( c )
+    return PL_unify_pointer(clause, c);
+  
   fail;
 }
