@@ -1,4 +1,4 @@
-/*  $Id: pl-stream.c,v 1.14 1996/08/16 12:00:29 jan Exp $
+/*  $Id: pl-stream.c,v 1.16 1996/10/02 16:33:15 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -95,7 +95,7 @@ S__removebuf(IOSTREAM *s)
 
     if ( !(s->flags & SIO_USERBUF) )
       free(s->buffer);
-    s->buffer = 0;
+    s->bufp = s->limitp = s->buffer = NULL;
     s->bufsize = 0;
 
     return rval;
@@ -198,6 +198,7 @@ S__fillbuf(IOSTREAM *s)
     if ( !s->buffer )
     { if ( S__setbuf(s, NULL, 0) < 0 )
 	return -1;
+      s->limitp = s->buffer;
     }
 
     if ( (n=(*ReadF(s))(s->handle, s->buffer, s->bufsize)) > 0 )
