@@ -1,4 +1,4 @@
-/*  $Id: pl-trace.c,v 1.28 1996/08/27 09:18:12 jan Exp $
+/*  $Id: pl-trace.c,v 1.32 1996/11/11 13:35:44 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -36,7 +36,7 @@ user to intercept and redefine the tracer.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 					/* Frame <-> Prolog integer */
-forwards LocalFrame	redoFrame(LocalFrame);
+forwards LocalFrame	redoFrame(LocalFrame, Code *PC);
 forwards int		traceAction(char *, int, LocalFrame, bool);
 forwards void		helpTrace(void);
 #ifdef O_INTERRUPT
@@ -45,7 +45,7 @@ forwards void		helpInterrupt(void);
 forwards bool		hasAlternativesFrame(LocalFrame);
 forwards void		alternatives(LocalFrame);
 forwards void		listProcedure(Definition);
-forwards int		traceInterception(LocalFrame, int);
+forwards int		traceInterception(LocalFrame, LocalFrame, int, Code);
 forwards bool		canUnifyTermWithGoal(Word, LocalFrame);
 forwards void		writeFrameGoal(LocalFrame frame, int how);
 
@@ -56,9 +56,11 @@ than the redo port of some subgoal of this port.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static LocalFrame
-redoFrame(register LocalFrame fr)
-{ for( ; fr && false(fr, FR_SKIPPED); fr = parentFrame(fr) )
-    ;
+redoFrame(LocalFrame fr, Code *PC)
+{ while( fr && false(fr, FR_SKIPPED))
+  { *PC = fr->programPointer;
+    fr = parentFrame(fr);
+  }
 
   return fr;
 }
@@ -107,7 +109,7 @@ returns to the WAM interpreter how to continue the execution:
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
-tracePort(LocalFrame frame, int port)
+tracePort(LocalFrame frame, LocalFrame bfr, int port, Code PC)
 { int OldOut;
   extern int Output;
   int action = ACTION_CONTINUE;
@@ -122,7 +124,8 @@ tracePort(LocalFrame frame, int port)
     Undo(frame->mark);
 
 					/* trace/[1,2] */
-  if ( true(def, TRACE_CALL|TRACE_REDO|TRACE_EXIT|TRACE_FAIL) )
+  if ( true(def, TRACE_CALL|TRACE_REDO|TRACE_EXIT|TRACE_FAIL) &&
+       !(port & (BREAK_PORT|CUT_PORT)) )
   { char *fmt = NULL;
 
     switch(port)
@@ -150,32 +153,39 @@ tracePort(LocalFrame frame, int port)
     }
   }
 
-  if ( (!debugstatus.tracing && false(def, SPY_ME))	|| /* non-tracing */
-       debugstatus.skiplevel < levelFrame(frame)	|| /* skipped */
-       false(def, TRACE_ME)				|| /* non-tracing */
-       (!(debugstatus.visible & port))			|| /* wrong port */
-       (port == REDO_PORT && (debugstatus.skiplevel == levelFrame(frame) ||
-			      (true(def, SYSTEM) && !SYSTEM_MODE)
-			     )) )				   /* redos */
+  if ( port != BREAK_PORT &&
+       ((!debugstatus.tracing && false(def, SPY_ME))	|| /* non-tracing */
+	debugstatus.skiplevel < levelFrame(frame)	|| /* skipped */
+	((port & (BREAK_PORT|CUT_PORT)) &&
+	 ((debugstatus.skiplevel == levelFrame(frame)) ||
+	  true(def, HIDE_CHILDS)))			|| /* also skipped */
+	false(def, TRACE_ME)				|| /* non-tracing */
+	(!(debugstatus.visible & port))			|| /* wrong port */
+	(port == REDO_PORT && (debugstatus.skiplevel == levelFrame(frame) ||
+			       (true(def, SYSTEM) && !SYSTEM_MODE)
+			      ))) )				   /* redos */
     return ACTION_CONTINUE;
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Give a trace on the skipped goal for a redo.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-  if ( port == REDO_PORT && debugstatus.skiplevel == VERY_DEEP &&
-       (fr = redoFrame(frame)) != NULL )
-  { debugstatus.skiplevel--;				   /* avoid a loop */
-    switch( tracePort(fr, REDO_PORT) )
-    { case ACTION_CONTINUE:
-	if ( debugstatus.skiplevel < levelFrame(frame) )
-	  return ACTION_CONTINUE;
-	break;
-      case ACTION_RETRY:
-      case ACTION_IGNORE:
-      case ACTION_FAIL:
-	Putf("Action not yet implemented here\n");
-	break;
+  { Code pc2;
+
+    if ( port == REDO_PORT && debugstatus.skiplevel == VERY_DEEP &&
+	 (fr = redoFrame(frame, &pc2)) != NULL )
+    { debugstatus.skiplevel--;				   /* avoid a loop */
+      switch( tracePort(fr, bfr, REDO_PORT, pc2) )
+      { case ACTION_CONTINUE:
+	  if ( debugstatus.skiplevel < levelFrame(frame) )
+	    return ACTION_CONTINUE;
+	  break;
+	case ACTION_RETRY:
+	case ACTION_IGNORE:
+	case ACTION_FAIL:
+	  Putf("Action not yet implemented here\n");
+	  break;
+      }
     }
   }
 
@@ -197,7 +207,8 @@ We are in searching mode; should we actually give this port?
 Do the Prolog trace interception.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-  if ((action = traceInterception(frame, port)) >= 0)
+  action = traceInterception(frame, bfr, port, PC);
+  if ( action >= 0 )
     return action;
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -218,6 +229,9 @@ again:
     case FAIL_PORT:	Putf(" Fail:  ");	break;
     case EXIT_PORT:	Putf(" Exit:  ");	break;
     case UNIFY_PORT:	Putf(" Unify: ");	break;
+    case BREAK_PORT:	Putf(" Break: ");	break;
+    case CUT_CALL_PORT:	Putf(" Cut call: ");	break;
+    case CUT_EXIT_PORT:	Putf(" Cut exit: ");	break;
   }
   Putf("(%3ld) ", levelFrame(frame));
   writeFrameGoal(frame, debugstatus.style);
@@ -629,7 +643,7 @@ backTrace(LocalFrame frame, int depth)
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Trace interception mechanism.  Whenever the tracer wants to perform some
 action   it   will   first   call   the    users'    Prolog    predicate
-prolog_trace_interception/3, allowing the user to define his/her action.
+prolog_trace_interception/4, allowing the user to define his/her action.
 If  this procedure succeeds the tracer assumes the trace action has been
 done and returns, otherwise the  default  C-defined  trace  actions  are
 performed.
@@ -643,38 +657,62 @@ This predicate is supposed to return one of the following atoms:
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
-traceInterception(LocalFrame frame, int port)
+traceInterception(LocalFrame frame, LocalFrame bfr, int port, Code PC)
 { int rval = -1;			/* Default C-action */
+  static predicate_t proc;
+
+  if ( !proc )
+    proc = PL_predicate("prolog_trace_interception", 4, "user");
+  if ( !proc->definition->definition.clauses )
+    return rval;
 
   if ( !status.boot && status.debugLevel == 0 )
   { fid_t cid = PL_open_foreign_frame();
     qid_t qid;
-    term_t argv = PL_new_term_refs(3);
-    atom_t portname;
-    static Procedure proc;
-
-    if ( !proc )
-      proc = lookupProcedure(FUNCTOR_traceinterc3, MODULE_user);
+    term_t argv = PL_new_term_refs(4);
+    term_t rarg = argv+3;
+    atom_t portname = NULL_ATOM;
+    functor_t portfunc = NULL;
 
     switch(port)
-    { case CALL_PORT:	portname = ATOM_call;	break;
-      case REDO_PORT:	portname = ATOM_redo;	break;
-      case EXIT_PORT:	portname = ATOM_exit;	break;
-      case FAIL_PORT:	portname = ATOM_fail;	break;
+    { case CALL_PORT:	  portname = ATOM_call;		break;
+      case REDO_PORT:	  portname = ATOM_redo;		break;
+      case EXIT_PORT:	  portname = ATOM_exit;		break;
+      case FAIL_PORT:	  portname = ATOM_fail;		break;
+      case UNIFY_PORT:	  portname = ATOM_unify;	break;
+      case BREAK_PORT:    portfunc = FUNCTOR_break1;	break;
+      case CUT_CALL_PORT: portfunc = FUNCTOR_cut_call1; break;
+      case CUT_EXIT_PORT: portfunc = FUNCTOR_cut_exit1; break;
       default:
-      case UNIFY_PORT:	portname = ATOM_unify;	break;
+	assert(0);
+        return rval;
+    }
+
+    if ( portname )
+      PL_put_atom(argv, portname);
+    else
+    { int pcn;
+
+      if ( PC && false(frame->predicate, FOREIGN) && frame->clause )
+	pcn = PC - frame->clause->clause->codes;
+      else
+	pcn = 0;
+
+      PL_unify_term(argv,
+		    PL_FUNCTOR, portfunc,
+		    PL_INTEGER, pcn);
     }
 
-    PL_put_atom(argv, portname);
     PL_put_integer(argv+1, PrologRef(frame));
-    PL_put_variable(argv+2);
+    PL_put_integer(argv+2, PrologRef(bfr));
+    PL_put_variable(rarg);
 
     debugstatus.suspendTrace++;
     qid = PL_open_query(MODULE_user, FALSE, proc, argv);
     if ( PL_next_solution(qid) )
     { atom_t a;
 
-      if ( PL_get_atom(argv+2, &a) )
+      if ( PL_get_atom(rarg, &a) )
       { if ( a == ATOM_continue )
 	  rval = ACTION_CONTINUE;
 	else if ( a == ATOM_fail )
@@ -683,6 +721,15 @@ traceInterception(LocalFrame frame, int port)
 	  rval = ACTION_RETRY;
 	else if ( a == ATOM_ignore )
 	  rval = ACTION_IGNORE;
+      } else if ( PL_is_functor(rarg, FUNCTOR_retry1) )
+      { long w;
+	term_t arg = PL_new_term_ref();
+
+	if ( PL_get_arg(1, rarg, arg) && PL_get_long(arg, &w) )
+	{ debugstatus.retryFrame = FrameRef(w);
+	  rval = ACTION_RETRY;
+	} else
+	  warning("prolog_trace_interception/3: bad argument to retry/1");
       }
     }
     PL_close_query(qid);
@@ -836,35 +883,73 @@ initTracer(void)
   pl_signal(SIGINT, interruptHandler);
 #endif
 
-  debugstatus.visible  = CALL_PORT|FAIL_PORT|REDO_PORT|EXIT_PORT;
-  debugstatus.leashing = CALL_PORT|FAIL_PORT|REDO_PORT|EXIT_PORT;
-  debugstatus.tracing = debugstatus.debugging = FALSE;
+  debugstatus.visible      = 
+  debugstatus.leashing     = CALL_PORT|FAIL_PORT|REDO_PORT|EXIT_PORT|
+			     BREAK_PORT;
+  debugstatus.tracing      =
+  debugstatus.debugging    = FALSE;
   debugstatus.suspendTrace = FALSE;
-  debugstatus.skiplevel = 0;
-  debugstatus.style = status.boot ? W_WRITE : W_PRINT; 
-  debugstatus.showContext = FALSE;
+  debugstatus.skiplevel    = 0;
+  debugstatus.style        = status.boot ? W_WRITE : W_PRINT; 
+  debugstatus.showContext  = FALSE;
+  debugstatus.retryFrame   = NULL;
 }
 
 		/********************************
 		*       PROLOG PREDICATES       *
 		*********************************/
 
-word
-pl_trace()
-{ debugstatus.debugging = debugstatus.tracing = TRUE;
-  debugstatus.skiplevel = VERY_DEEP;
-  find.searching = FALSE;
+int
+tracemode(int doit, int *old)
+{ if ( doit )
+    doit = TRUE;
+
+  if ( old )
+    *old = debugstatus.tracing;
+
+  if ( debugstatus.tracing != doit )
+  { if ( doit )
+    { debugstatus.debugging = TRUE;
+      debugstatus.skiplevel = VERY_DEEP;
+      find.searching = FALSE;
+    }
+    debugstatus.tracing = doit;
+    callEventHook(PLEV_TRACING, doit);
+  }
 
   succeed;
 }
 
-word
-pl_notrace()
-{ debugstatus.tracing = FALSE;
+
+int
+debugmode(int doit, int *old)
+{ if ( doit )
+    doit = TRUE;
+
+  if ( old )
+    *old = debugstatus.debugging;
+
+  if ( debugstatus.debugging != doit )
+  { if ( doit )
+      debugstatus.skiplevel = VERY_DEEP;
+    debugstatus.debugging = doit;
+    callEventHook(PLEV_DEBUGGING, doit);
+  }
 
   succeed;
 }
 
+
+word
+pl_trace()
+{ return tracemode(TRUE, NULL);
+}
+
+word
+pl_notrace()
+{ return tracemode(FALSE, NULL);
+}
+
 word
 pl_tracing()
 { return debugstatus.tracing;
@@ -872,25 +957,17 @@ pl_tracing()
 
 word
 pl_debug()
-{ debugstatus.debugging = TRUE;
-  debugstatus.skiplevel = VERY_DEEP;
-
-  succeed;
+{ return debugmode(TRUE, NULL);
 }
 
 word
 pl_nodebug()
-{ debugstatus.debugging = FALSE;
-
-  succeed;
+{ return debugmode(FALSE, NULL);
 }
 
 word
 pl_debugging()
-{ if ( debugstatus.debugging )
-    succeed;
-
-  fail;
+{ return debugstatus.debugging;
 }
 
 word
@@ -991,20 +1068,56 @@ pl_prolog_frame_attribute(term_t frame, term_t what,
 			  term_t value)
 { LocalFrame fr;
   atom_t key;
+  int arity;
   term_t result = PL_new_term_ref();
   long fri;
 
   if ( !PL_get_long(frame, &fri) ||
-       !PL_get_atom(what, &key) )
+       !PL_get_name_arity(what, &key, &arity) )
+  { ierr:
     return warning("prolog_frame_attribute/3: instantiation fault");
+  }
 
   if ((fr = FrameRef(fri)) < lBase || fr > lTop)
     return warning("prolog_frame_attribute/3: illegal frame reference");
 
+  if ( key == ATOM_argument && arity == 1 )
+  { term_t arg = PL_new_term_ref();
+    int argn;
+    Word p = valTermRef(value);
+
+    if ( !PL_get_arg(1, what, arg) || !PL_get_integer(arg, &argn) || argn < 1 )
+      goto ierr;
+
+    if ( true(fr->predicate, FOREIGN) || !fr->clause )
+    { if ( argn > fr->predicate->functor->arity )
+	fail;
+    } else
+    { if ( argn > fr->clause->clause->prolog_vars )
+	fail;
+    }
+
+#ifdef O_DEBUGLOCAL			/* see pl-wam.c */
+    assert( *argFrameP(fr, argn-1) != (word)(((char*)ATOM_nil) + 1) );
+    checkData(argFrameP(fr, argn-1), FALSE);
+#endif
+
+   deRef(p);
+   if ( isVar(*p) )
+   { *p = makeRef(argFrameP(fr, argn-1));
+     DoTrail(p);
+     succeed;
+   }
+
+   fail;
+  }
+  if ( arity != 0 )
+    goto ierr;
+
   if (        key == ATOM_level)
   { PL_put_integer(result, levelFrame(fr));
   } else if (key == ATOM_has_alternatives)
-  { PL_put_integer(result, hasAlternativesFrame(fr));
+  { PL_put_atom(result, hasAlternativesFrame(fr) ? ATOM_true : ATOM_false);
   } else if (key == ATOM_alternative)
   { if (fr->backtrackFrame == (LocalFrame) NULL)
       fail;
@@ -1012,17 +1125,20 @@ pl_prolog_frame_attribute(term_t frame, term_t what,
   } else if (key == ATOM_parent)
   { LocalFrame parent;
 
+    if ( fr->parent )
+      clearUninitialisedVarsFrame(fr->parent, fr->programPointer);
+
     if ( (parent = parentFrame(fr)) )
       PL_put_integer(result, PrologRef(parent));
     else
       fail;
   } else if (key == ATOM_top)
-  { PL_put_integer(result, fr->parent ? 1 : 0);
+  { PL_put_atom(result, fr->parent ? ATOM_false : ATOM_true);
   } else if (key == ATOM_context_module)
   { PL_put_atom(result, contextModule(fr)->name);
   } else if (key == ATOM_clause)
   { if ( false(fr->predicate, FOREIGN) && fr->clause )
-      PL_put_pointer(result, fr->clause);
+      PL_put_pointer(result, fr->clause->clause);
     else
       fail;
   } else if (key == ATOM_goal)
@@ -1048,8 +1164,91 @@ pl_prolog_frame_attribute(term_t frame, term_t what,
 	unify_ptrs(valTermRef(a), argFrameP(fr, n));
       }
     }
+  } else if ( key == ATOM_pc )
+  { if ( fr->programPointer &&
+	 fr->parent &&
+	 false(fr->parent->predicate, FOREIGN) )
+      PL_put_integer(result,
+		     fr->programPointer - fr->parent->clause->clause->codes);
+    else
+      fail;
+  } else if ( key == ATOM_hidden )
+  { atom_t a;
+
+    if ( SYSTEM_MODE )
+    { a = ATOM_true;
+    } else
+    { if ( true(fr, FR_NODEBUG) || false(fr->predicate, TRACE_ME) )
+	a = ATOM_true;
+      else
+	a = ATOM_false;
+    }
+
+    PL_put_atom(result, a);
   } else
     return warning("prolog_frame_attribute/3: unknown key");
 
   return PL_unify(value, result);
 }
+
+
+		 /*******************************
+		 *	  PROLOG EVENET HOOK	*
+		 *******************************/
+
+void
+callEventHook(int ev, ...)
+{ if ( !PROCEDURE_event_hook1 )
+    PROCEDURE_event_hook1 = PL_predicate("prolog_event_hook", 1, "user");
+  
+  if ( PROCEDURE_event_hook1->definition->definition.clauses )
+  { va_list args;
+    fid_t fid = PL_open_foreign_frame();
+    term_t arg = PL_new_term_ref();
+
+    va_start(args, ev);
+    switch(ev)
+    { case PLEV_ERASED:
+      {	void *ptr = va_arg(args, void *); 	/* object erased */
+
+	PL_unify_term(arg, PL_FUNCTOR, FUNCTOR_erased1,
+		           PL_POINTER, ptr);
+	break;
+      }
+      case PLEV_DEBUGGING:
+      { int dbg = va_arg(args, int);
+	
+	PL_unify_term(arg, PL_FUNCTOR, FUNCTOR_debugging1,
+			   PL_ATOM, dbg ? ATOM_true : ATOM_false);
+	break;
+      }
+      case PLEV_TRACING:
+      { int trc = va_arg(args, int);
+	
+	PL_unify_term(arg, PL_FUNCTOR, FUNCTOR_tracing1,
+			   PL_ATOM, trc ? ATOM_true : ATOM_false);
+	break;
+      }
+      case PLEV_BREAK:
+      case PLEV_NOBREAK:
+      { Clause clause = va_arg(args, Clause);
+	int offset = va_arg(args, int);
+
+	PL_unify_term(arg, PL_FUNCTOR, FUNCTOR_break3,
+		           PL_POINTER, clause,
+		           PL_INTEGER, offset,
+			   PL_ATOM, ev == PLEV_BREAK ? ATOM_true
+					             : ATOM_false);
+	break;
+      }
+      default:
+	warning("callEventHook(): unknown event: %d", ev);
+        goto out;
+    }
+    
+    PL_call_predicate(MODULE_user, FALSE, PROCEDURE_event_hook1, arg);
+  out:
+    PL_discard_foreign_frame(fid);
+    va_end(args);
+  }
+}
