@@ -1,4 +1,4 @@
-/*  $Id: pl-wam.c,v 1.59 1996/09/06 09:34:11 jan Exp $
+/*  $Id: pl-wam.c,v 1.62 1996/10/30 17:33:11 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -67,6 +67,8 @@ struct
   int i_depart;
   int i_call;
   int i_exit;
+  int i_exitfact;
+  int d_break;
 #if O_COMPILE_ARITH
   int a_indirect;
   int a_func0[256];
@@ -129,6 +131,8 @@ pl_count()
   countOne(  "I_DEPART", 	counting.i_depart);
   countOne(  "I_CALL", 		counting.i_call);
   countOne(  "I_EXIT", 		counting.i_exit);
+  countOne(  "I_EXITFACT",	counting.i_exitfact);
+  countOne(  "D_BREAK",		countOne.d_break);
   countOne(  "I_FAIL",		countOne.i_fail);
   countOne(  "I_TRUE",		countOne.i_true);
 
@@ -657,7 +661,8 @@ right_recursion:
   { case TAG_ATOM:
       fail;
     case TAG_INTEGER:
-      if ( storage(w1) == STG_INLINE )
+      if ( storage(w1) == STG_INLINE ||
+	   storage(w2) == STG_INLINE )
 	fail;
     case TAG_STRING:
     case TAG_FLOAT:
@@ -1148,18 +1153,22 @@ pl-comp.c
     &&C_SOFTIF_LBL,
     &&C_SOFTCUT_LBL,
 #endif
+    &&I_EXITFACT_LBL,
+    &&D_BREAK_LBL,
     NULL
   };
 
 #define VMI(Name, Count, Msg)	Name ## _LBL: Count; DEBUG(8, Sdprintf Msg);
 #if VMCODE_IS_ADDRESS
-#define NEXT_INSTRUCTION	goto *(void *)((int)(*PC++))
+#define NEXT_INSTRUCTION	goto *(void *)((long)(*PC++))
 #else
 #define NEXT_INSTRUCTION	goto *jmp_table[*PC++]
 #endif
 
 #else /* O_LABEL_ADDRESSES */
 
+code thiscode;
+
 #define VMI(Name, Count, Msg)	case Name: Count; DEBUG(8, Sdprintf Msg);
 #define NEXT_INSTRUCTION	goto next_instruction
 
@@ -1193,7 +1202,7 @@ depart_continue() to do the normal thing or to the backtrack point.
     { Undo(FR->mark);
 #if O_DEBUGGER
       if ( debugstatus.debugging )
-      { switch( tracePort(FR, REDO_PORT) )
+      { switch( tracePort(FR, BFR, REDO_PORT, NULL) )
 	{ case ACTION_FAIL:
 	    QF->deterministic = TRUE;
 	    fail;
@@ -1226,9 +1235,49 @@ initialised properly.
   NEXT_INSTRUCTION;
 #else
 next_instruction:
-  switch( *PC++ )
+  thiscode = *PC++;
+resumebreak:
+  switch( thiscode )
 #endif
-  { VMI(I_NOP, COUNT(i_nop), ("i_nop\n"))
+  {
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+D_BREAK implements break-points in the  code.   A  break-point is set by
+replacing  an  instruction  by  a   D_BREAK  instruction.  The  orininal
+instruction is saved in a table. replacedBreak() fetches it.
+
+We might be in a state where  we   are  writing  the arguments above the
+current lTop, and therefore with higher this  with the maximum number of
+arguments.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+    VMI(D_BREAK, COUNT(d_break), ("d_break\n"))
+#if O_DEBUGGER
+    if ( debugstatus.debugging )
+    { int action;
+      LocalFrame lSave = lTop;
+
+      lTop = (LocalFrame)argFrameP(lTop, MAXARITY);
+      clearUninitialisedVarsFrame(FR, PC-1);
+      action = tracePort(FR, BFR, BREAK_PORT, PC-1);
+      lTop = lSave;
+
+      switch(action)
+      { case ACTION_RETRY:
+	  goto retry;
+      }
+    }
+#endif /*O_DEBUGGER*/
+#ifdef O_LABEL_ADDRESSES
+    { void *c = (void *)replacedBreak(PC-1);
+      
+      goto *c;
+    }
+#else
+    thiscode = replacedBreak(PC-1);
+    goto resumebreak;
+#endif      
+
+    VMI(I_NOP, COUNT(i_nop), ("i_nop\n"))
 	NEXT_INSTRUCTION;
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -1682,11 +1731,16 @@ makes debugging much more  difficult  as  the  system  will  do  a  deep
 backtrack without showing the fail ports explicitely.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
     VMI(I_ENTER, COUNT(i_enter), ("enter\n")) MARK(ENTER);
-      { ARGP = argFrameP(lTop, 0);
-
+      { 
 #if O_DEBUGGER
 	if ( debugstatus.debugging )
-	{ tracePort(FR, UNIFY_PORT);
+	{ clearUninitialisedVarsFrame(FR, PC);
+	  switch(tracePort(FR, BFR, UNIFY_PORT, PC))
+	  { case ACTION_RETRY:
+	      goto retry;
+	    case ACTION_FAIL:
+	      FRAME_FAILED;
+	  }
 	  if ( FR->mark.trailtop == INVALID_TRAILTOP )
 	  { SetBfr(FR->backtrackFrame);
 	  } else
@@ -1701,7 +1755,8 @@ backtrack without showing the fail ports explicitely.
 	  }
 	}
 
-	NEXT_INSTRUCTION;
+	ARGP = argFrameP(lTop, 0);
+        NEXT_INSTRUCTION;
       }
 
 #if O_BLOCK
@@ -1836,6 +1891,17 @@ backtrack that makes it difficult to understand the tracer's output.
       { LocalFrame fr;
 	register LocalFrame fr2;
 
+#ifdef O_DEBUGGER
+	if ( debugstatus.debugging )
+	{ switch(tracePort(FR, BFR, CUT_CALL_PORT, PC))
+	  { case ACTION_RETRY:
+	      goto retry;
+	    case ACTION_FAIL:
+	      FRAME_FAILED;
+	  }
+	}
+#endif
+
 	set(FR, FR_CUT);
 	for(fr = BFR; fr > FR; fr = fr->backtrackFrame)
 	{ for(fr2 = fr; fr2->clause && fr2 > FR; fr2 = fr2->parent)
@@ -1858,6 +1924,17 @@ backtrack that makes it difficult to understand the tracer's output.
 	lTop = (LocalFrame) argFrameP(FR, CL->clause->variables);
 	ARGP = argFrameP(lTop, 0);
 
+#ifdef O_DEBUGGER
+	if ( debugstatus.debugging )
+	{ switch(tracePort(FR, BFR, CUT_EXIT_PORT, PC))
+	  { case ACTION_RETRY:
+	      goto retry;
+	    case ACTION_FAIL:
+	      FRAME_FAILED;
+	  }
+	}
+#endif
+
 	NEXT_INSTRUCTION;
       }
 
@@ -2796,6 +2873,9 @@ Note: we are working above `lTop' here!
 	}
 #endif
 	incLevel(FR);
+#ifdef O_DEBUGGER
+      retry_continue:
+#endif
 	clear(FR, FR_CUT|FR_SKIPPED);
 
 	statistics.inferences++;
@@ -2879,9 +2959,10 @@ Testing is suffices to find out that the predicate is defined.
 	if ( debugstatus.debugging )
 	{ lTop = (LocalFrame) argFrameP(FR, DEF->functor->arity);
 	  CL = DEF->definition.clauses;
-	  switch(tracePort(FR, CALL_PORT))
+	  switch(tracePort(FR, BFR, CALL_PORT, NULL))
 	  { case ACTION_FAIL:	goto frame_failed;
 	    case ACTION_IGNORE: goto exit_builtin;
+	    case ACTION_RETRY:  goto retry;
 	  }
 	}
 #endif /*O_DEBUGGER*/
@@ -2978,6 +3059,23 @@ Leave the clause:
 	}
 	goto normal_exit;
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+i_exitfact is generated to close a fact. The reason for not generating a
+plain I_EXIT is first of all that the actual sequence should be I_ENTER,
+I_EXIT,  and  just  optimising   to    I_EXIT   looses   the  unify-port
+interception. Second, there should be some room for optimisation here.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+    VMI(I_EXITFACT, COUNT(i_exitfact), ("exitfact ")) MARK(EXITFACT);
+#if O_DEBUGGER
+	if ( debugstatus.debugging )
+	{ switch(tracePort(FR, BFR, UNIFY_PORT, PC))
+	  { case ACTION_RETRY:
+	      goto retry;
+	  }
+	}
+#endif /*O_DEBUGGER*/
+	/*FALLTHROUGH*/
+
     VMI(I_EXIT, COUNT(i_exit), ("exit ")) MARK(EXIT);
 	if ( false(FR, FR_CUT) )
 	{ if ( FR > BFR )			/* alternatives */
@@ -2987,30 +3085,41 @@ Leave the clause:
 	{ if ( BFR <= FR )			/* deterministic */
 	  { if ( BFR == FR )
 	      SetBfr(FR->backtrackFrame);
-	    lTop = FR;
 	    leaveDefinition(DEF);
+	    lTop = FR;
 	  }
 	  deterministic = TRUE;
 	}
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+First, call the tracer. Basically,  the   current  frame is garbage, but
+given that the tracer might need to print the variables, we have to be a
+bit more careful.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
     normal_exit:
 #if O_DEBUGGER
 	if ( debugstatus.debugging )
-	{ int action;
+        { LocalFrame mintop;
+	  int action;
+
 	  LocalFrame lSave = lTop;
 	  environment_frame = FR;
 
-	  if ( lTop < (LocalFrame)argFrameP(FR, DEF->functor->arity) )
-	    lTop = (LocalFrame)argFrameP(FR, DEF->functor->arity);
-	  action = tracePort(FR, EXIT_PORT);
+	  if ( false(DEF, FOREIGN) && CL )
+	    mintop = (LocalFrame) argFrameP(FR, CL->clause->variables);
+	  else
+	    mintop = (LocalFrame) argFrameP(FR, DEF->functor->arity);
 
-	  switch(action)
+	  if ( lTop < mintop )
+	    lTop = mintop;
+
+	  action = tracePort(FR, BFR, EXIT_PORT, PC);
+	  lTop = lSave;
+	  switch( action )
 	  { case ACTION_RETRY:	goto retry;
 	    case ACTION_FAIL:	set(FR, FR_CUT);
 				FRAME_FAILED;
 	  }
-
-	  lTop = lSave;
 	}
 #endif /*O_DEBUGGER*/
 
@@ -3051,41 +3160,49 @@ Leave the clause:
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 			TRACER RETRY ACTION
 
-To retry we should first undo all actions done since the start  of  this
-frame  by  resetting  the  global  stack and calling Undo(). The current
-frame becomes the backtrack frame for the new childs.
+By default, retries the  current  frame.  If   another  frame  is  to be
+retried, place the frame-reference, which  should   be  a  parent of the
+current frame, in debugstatus.retryFrame and jump to this label. This is
+implemented by returning retry(Frame) of the prolog_trace_interception/3
+hook.
 
-Foreign functions can now just be restarted.  For Prolog  ones  we  will
-create  a  dummy  clause before the first one and proceed as with normal
-backtracking.
-
-BUG: Clause reference counts should be  updated  properly.   Needs  some
-detailed study!
+First, the system will leave any parent  frames. Next, it will undo back
+to the call-port and finally, restart the clause.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#if O_DEBUGGER
 retry:					MARK(RETRY);
+{ LocalFrame rframe = debugstatus.retryFrame;
+  LocalFrame fr;
+
+  if ( !rframe )
+    rframe = FR;
+  debugstatus.retryFrame = NULL;
+
+  Sdprintf("[Retrying frame %d running %s]\n",
+	   (Word)rframe - (Word)lBase,
+	   predicateName(rframe->predicate));
+
+  for(fr = BFR; fr > rframe; fr = fr->backtrackFrame)
+  { LocalFrame fr2;
+
+    for(fr2 = fr; fr2->clause && fr2 > rframe; fr2 = fr2->parent)
+    { DEBUG(3, Sdprintf("discard %d\n", (Word)fr2 - (Word)lBase) );
+      leaveFrame(fr2);
+      fr2->clause = NULL;
+    }
+  }
+
+  environment_frame = FR = rframe;
+  DEF = FR->predicate;
   Undo(FR->mark);
   SetBfr(FR);
   clear(FR, FR_CUT);
   lTop = (LocalFrame) argFrameP(FR, DEF->functor->arity);
-#if O_DEBUGGER
-  if ( debugstatus.debugging )
-  { tracePort(FR, CALL_PORT);
-  }
+
+  goto retry_continue;
+}
 #endif /*O_DEBUGGER*/
-  if ( false(DEF, FOREIGN) )
-  { struct clause_ref zero;		/* fake a clause */
-    struct clause zclause;
-
-    clear(&zclause, ERASED);		/* avoid destruction */
-    zero.next   = DEF->definition.clauses;
-    zero.clause = &zclause;
-    CL = &zero;
-
-    CLAUSE_FAILED;
-  }
-  FR->clause = FIRST_CALL;
-  goto call_builtin;
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 The rest of this giant procedure handles backtracking.  There are  three
@@ -3096,7 +3213,7 @@ different ways we can get here:
     procedure  and  if we are out of clauses continue with the backtrack
     frame of this frame.
 
-  - A foreign goal failed				(frame_failed)
+<  - A foreign goal failed				(frame_failed)
     In this case we can continue at the backtrack frame of  the  current
     frame.
 
@@ -3184,7 +3301,7 @@ frame_failed:				MARK(FAIL);
 
 #if O_DEBUGGER
     if ( debugstatus.debugging )
-    { switch( tracePort(FR, FAIL_PORT) )
+    { switch( tracePort(FR, BFR, FAIL_PORT, PC) )
       { case ACTION_RETRY:	goto retry;
 	case ACTION_IGNORE:	Putf("ignore not (yet) implemented here\n");
       }
@@ -3240,7 +3357,7 @@ resume_from_body:
     { Undo(FR->mark);			/* data backtracking to get nice */
 					/* tracer output */
 
-      switch( tracePort(FR, REDO_PORT) )
+      switch( tracePort(FR, BFR, REDO_PORT, NULL) )
       { case ACTION_FAIL:	continue;
 	case ACTION_IGNORE:	CL = NULL;
 				goto exit_builtin;
