@@ -1,4 +1,4 @@
-/*  $Id: pl-wic.c,v 1.49 1996/09/06 09:34:15 jan Exp $
+/*  $Id: pl-wic.c,v 1.51 1996/10/30 17:33:15 jan Exp $
 
     Copyright (c) 1990 Jan Wielemaker. All rights reserved.
     See ../LICENCE to find out about your rights.
@@ -676,7 +676,7 @@ loadWicFd(char *file, IOSTREAM *fd, int flags)
 
 					/* fix paths for changed home */
   savedhome = getString(fd);
-  if ( streq(savedhome, systemDefaults.home) )
+  if ( !systemDefaults.home || streq(savedhome, systemDefaults.home) )
   { qlf_has_moved = FALSE;
   } else
   { qlf_has_moved = TRUE;
@@ -816,7 +816,8 @@ loadPredicate(IOSTREAM *fd, int skip)
 	clearFlags(clause);
 	clause->prolog_vars = (short) getNum(fd);
 	clause->variables = (short) getNum(fd);
-	clause->subclauses = (short) getNum(fd);
+	if ( getNum(fd) == 0 )		/* 0: fact */
+	  set(clause, UNIT_CLAUSE);
 	clause->procedure = proc;
 	clause->source_no = (currentSource ? currentSource->index : 0);
 	clause->code_size = (short) getNum(fd);
@@ -1345,7 +1346,7 @@ saveWicClause(Clause clause, IOSTREAM *fd)
   putNum(clause->line_no, fd);
   putNum(clause->prolog_vars, fd);
   putNum(clause->variables, fd);
-  putNum(clause->subclauses, fd);
+  putNum(true(clause, UNIT_CLAUSE) ? 0 : 1, fd);
   putNum(clause->code_size, fd);
 
   bp = clause->codes;
@@ -1535,7 +1536,10 @@ openWic(const char *file, term_t args)
   putString(goal,          	  wicFd);
   putString(topLevel,      	  wicFd);
   putString(initFile, 	   	  wicFd);
-  putString(systemDefaults.home,  wicFd);
+  if ( systemDefaults.home )
+    putString(systemDefaults.home,  wicFd);
+  else
+    putString("<no home>",  wicFd);
 
   currentProc    = (Procedure) NULL;
   currentSource  = (SourceFile) NULL;
